import{d as _,u as f,c as a,a as u,w as c,Q as k,r as w,o,j as y,e,R as Q,h as v,S as x,T as $,F as g,g as C}from"./index.4d314fbd.js";const q=_({__name:"WorkQueues",setup(b){const p={interval:3e4},s=f(k.getWorkQueues,[{}],p),r=a(()=>{var n;return(n=s.response)!=null?n:[]}),l=a(()=>s.executed&&r.value.length==0),d=a(()=>s.executed);return(n,t)=>{const i=w("p-layout-default");return o(),u(i,{class:"queues"},{header:c(()=>[y(e(Q))]),default:c(()=>[e(d)?(o(),v(g,{key:0},[e(l)?(o(),u(e(x),{key:0})):(o(),u(e($),{key:1,"work-queues":e(r),onUpdate:t[0]||(t[0]=m=>e(s).refresh()),onDelete:t[1]||(t[1]=m=>e(s).refresh())},null,8,["work-queues"]))],64)):C("",!0)]),_:1})}}});export{q as default};
