import{d as p,a as i,w as a,b as d,r as f,o as m,j as s,e as t,U as w,V as h,q as u,Q as k,W as c}from"./index.4d314fbd.js";const y=p({__name:"WorkQueueCreate",setup(Q){const o=d(),n=()=>{o.push(u.workQueues())},_=async r=>{try{const{id:e}=await k.createWorkQueue(r);c("Work queue has been created","success"),o.push(u.workQueue(e))}catch(e){c("Error occurred while creating new work queue","error"),console.error(e)}};return(r,e)=>{const l=f("p-layout-default");return m(),i(l,null,{header:a(()=>[s(t(w))]),default:a(()=>[s(t(h),{onSubmit:_,onCancel:n})]),_:1})}}});export{y as default};
