import{d as p,a7 as _,a as m,w as n,r as d,o as C,j as e,e as i,as as h,at as w,ao as y,Z as s,q as r,W as v}from"./index.4d314fbd.js";const b=p({__name:"NotificationCreate",setup(x){const a=_({});async function c(o){try{await y.createNotification(o),s.push(r.notifications())}catch(t){v("Error creating notification","error"),console.warn(t)}}function f(){s.push(r.notifications())}return(o,t)=>{const u=d("p-layout-default");return C(),m(u,null,{header:n(()=>[e(i(h))]),default:n(()=>[e(i(w),{notification:a.value,"onUpdate:notification":t[0]||(t[0]=l=>a.value=l),onSubmit:c,onCancel:f},null,8,["notification"])]),_:1})}}});export{b as default};
