import{d as y,c as b,o as i,h as _,j as t,w as n,F as x,av as C,aw as m,e as a,ax as w,ay as B,az as h,aA as $,X as v,a7 as E,aB as F,aC as M,a as O,r,aD as V,aE as j,aF as D}from"./index.4d314fbd.js";const X={class:"settings-block"},z=y({__name:"SettingsCodeBlock",props:{engineSettings:null},setup(c){const e=c,s=b(()=>Object.entries(e.engineSettings));return(u,d)=>(i(),_("div",X,[t(a(w),{multiline:""},{default:n(()=>[(i(!0),_(x,null,C(a(s),(l,o)=>(i(),_("div",{key:o,class:"settings-block--code-line"},m(l[0])+": "+m(l[1]),1))),128))]),_:1})]))}});class L extends h($){}const f=B(new L);const U=y({__name:"Settings",async setup(c){let e,s;const u=[{text:"Settings"}],d=([e,s]=v(()=>f.getSettings()),e=await e,s(),e),l=([e,s]=v(()=>f.getVersion()),e=await e,s(),e),o=E(F());return M(()=>D(o.value)),(N,p)=>{const S=r("p-key-value"),g=r("p-label"),k=r("p-layout-default");return i(),O(k,{class:"settings"},{header:n(()=>[t(a(V),{crumbs:u},{actions:n(()=>[t(S,{label:"Version",value:a(l),alternate:""},null,8,["value"])]),_:1})]),default:n(()=>[t(g,{label:"Color Mode",class:"settings__color-mode"},{default:n(()=>[t(a(j),{selected:o.value,"onUpdate:selected":p[0]||(p[0]=A=>o.value=A)},null,8,["selected"])]),_:1}),t(g,{label:"Orion Settings"},{default:n(()=>[t(z,{class:"settings__code-block","engine-settings":a(d)},null,8,["engine-settings"])]),_:1})]),_:1})}}});export{U as default};
