import{d as x,c as l,u as c,C as F,a,w as t,b as g,H as S,r as m,o as r,e,I as A,g as u,J as w,j as d,K as B,z as I,q as N}from"./index.4d314fbd.js";import{u as R}from"./useRouteParam.0009ec42.js";const z=x({__name:"Flow",setup(V){const f=R("id"),_=g(),y=l(()=>{const s=["Deployments"];return I.xl||s.unshift("Details"),s}),i={interval:3e5},b=c(F.getFlow,[f.value],i),o=l(()=>b.response),D=l(()=>({flows:{id:{any_:[f.value]}}})),p=c(S.getDeployments,[D],i),v=l(()=>{var s;return(s=p.response)!=null?s:[]});function k(){_.push(N.flows())}return(s,n)=>{const C=m("p-tabs"),$=m("p-layout-well");return r(),a($,{class:"flow"},{header:t(()=>[e(o)?(r(),a(e(A),{key:0,flow:e(o),onDelete:k},null,8,["flow"])):u("",!0)]),well:t(()=>[e(o)?(r(),a(e(w),{key:0,flow:e(o)},null,8,["flow"])):u("",!0)]),default:t(()=>[d(C,{tabs:e(y)},{details:t(()=>[e(o)?(r(),a(e(w),{key:0,flow:e(o)},null,8,["flow"])):u("",!0)]),deployments:t(()=>[d(e(B),{deployments:e(v),onUpdate:n[0]||(n[0]=h=>e(p).refresh()),onDelete:n[1]||(n[1]=h=>e(p).refresh())},null,8,["deployments"])]),_:1},8,["tabs"])]),_:1})}}});export{z as default};
