import{d as f,u as i,f as p,c as d,a,w as s,b as w,r as m,o as u,e as o,_ as R,g as v,j as h,B as y,q as g}from"./index.4d314fbd.js";import{u as x}from"./useRouteParam.0009ec42.js";const D=f({__name:"FlowRunRadar",setup(C){const t=w(),e=x("id"),r={interval:5e3},l=i(p.getFlowRun,[e.value],r),n=d(()=>l.response);function _(){t.push(g.flowRuns())}return(F,b)=>{const c=m("p-layout-full");return u(),a(c,{class:"flow-run-radar"},{header:s(()=>[o(n)?(u(),a(o(R),{key:0,"flow-run":o(n),class:"flow-run-radar__header",onDelete:_},null,8,["flow-run"])):v("",!0)]),default:s(()=>[h(o(y),{"flow-run-id":o(e)},null,8,["flow-run-id"])]),_:1})}}});export{D as default};
