import{d as v,c as d,u as g,a as o,w as l,b as h,f as F,r as _,o as n,e as a,_ as x,g as t,h as P,i as B,j as s,k as C,l as D,m as I,n as N,P as i,p as m,q as f,s as w,F as S,t as T,v as V,x as j,y as q,z}from"./index.4d314fbd.js";import{u as A}from"./useRouteParam.0009ec42.js";const E={class:"flow-run__meta"},K=v({__name:"FlowRun",setup(L){const p=h(),r=A("id"),k=d(()=>{const u=["Logs","Task Runs","Sub Flow Runs","Parameters"];return z.xl||u.push("Details"),u}),y=g(F.getFlowRun,[r],{interval:5e3}),e=d(()=>y.response);function R(){p.push(f.flowRuns())}return(u,G)=>{const c=_("router-link"),$=_("p-tabs"),b=_("p-layout-well");return n(),o(b,{class:"flow-run"},{header:l(()=>[a(e)?(n(),o(a(x),{key:0,"flow-run":a(e),onDelete:R},null,8,["flow-run"])):t("",!0)]),well:l(()=>[a(e)?(n(),P(S,{key:0},[B("div",E,[s(a(C),{state:a(e).state},null,8,["state"]),s(a(D),{duration:a(e).duration},null,8,["duration"]),s(a(I),{"flow-id":a(e).flowId},null,8,["flow-id"]),a(e).deploymentId?(n(),o(a(N),{key:0,"deployment-id":a(e).deploymentId},null,8,["deployment-id"])):t("",!0)]),s(a(i)),s(c,{to:a(f).radar(a(r)),class:"flow-run__small-radar-link"},{default:l(()=>[s(a(m),{"flow-run-id":a(r),class:"flow-run__small-radar"},null,8,["flow-run-id"])]),_:1},8,["to"]),s(a(i)),s(a(w),{"flow-run":a(e),alternate:""},null,8,["flow-run"])],64)):t("",!0)]),default:l(()=>[s($,{tabs:a(k)},{details:l(()=>[s(c,{to:a(f).radar(a(r)),class:"flow-run__small-radar-link"},{default:l(()=>[s(a(m),{"flow-run-id":a(r),class:"flow-run__small-radar"},null,8,["flow-run-id"])]),_:1},8,["to"]),s(a(i)),a(e)?(n(),o(a(w),{key:0,"flow-run":a(e)},null,8,["flow-run"])):t("",!0)]),logs:l(()=>[a(e)?(n(),o(a(T),{key:0,"flow-run":a(e)},null,8,["flow-run"])):t("",!0)]),"task-runs":l(()=>[a(e)?(n(),o(a(V),{key:0,"flow-run-id":a(e).id},null,8,["flow-run-id"])):t("",!0)]),"sub-flow-runs":l(()=>[a(e)?(n(),o(a(j),{key:0,"flow-run-id":a(e).id},null,8,["flow-run-id"])):t("",!0)]),parameters:l(()=>[a(e)?(n(),o(a(q),{key:0,parameters:a(e).parameters},null,8,["parameters"])):t("",!0)]),_:1},8,["tabs"])]),_:1})}}});export{K as default};
