import{d as _,a7 as b,c as o,u as f,a as k,w as l,a8 as y,r as m,o as d,j as c,e as s,a9 as v,aa as B}from"./index.4d314fbd.js";const g=_({__name:"BlocksCatalog",setup(C){const a=b(null),n=o(()=>a.value?{blockSchemas:{blockCapabilities:{all_:[a.value]}}}:{}),p=f(y.getBlockTypes,[n]),i=o(()=>{var e;return(e=p.response)!=null?e:[]});return(e,t)=>{const r=m("p-layout-default");return d(),k(r,{class:"blocks-catalog"},{header:l(()=>[c(s(v))]),default:l(()=>[c(s(B),{capability:a.value,"onUpdate:capability":t[0]||(t[0]=u=>a.value=u),"block-types":s(i)},null,8,["capability","block-types"])]),_:1})}}});export{g as default};
