import{d as Q,c as r,u as g,Q as h,a as t,w as o,b as v,r as l,o as s,e,$ as C,g as a,j as c,a0 as $,a1 as p,a2 as x,z as W,q as U}from"./index.4d314fbd.js";import{u as A}from"./useRouteParam.0009ec42.js";const S=Q({__name:"WorkQueue",setup(B){const k=v(),m=r(()=>{const n=["Upcoming runs"];return W.xl||n.unshift("Details"),n}),_=A("id"),d=r(()=>`prefect agent start ${_.value}`),w={interval:3e5},i=g(h.getWorkQueue,[_.value],w),u=r(()=>i.response),f=()=>{k.push(U.workQueues())};return(n,D)=>{const q=l("p-tabs"),b=l("p-layout-well"),y=l("p-layout-default");return s(),t(y,{class:"work-queue"},{header:o(()=>[e(u)?(s(),t(e(C),{key:0,queue:e(u),onUpdate:e(i).refresh,onDelete:f},null,8,["queue","onUpdate"])):a("",!0)]),default:o(()=>[c(b,{class:"work-queue__body"},{header:o(()=>[c(e($),{command:e(d),title:"Work queue is ready to go!",subtitle:"Work queues define the work to be done and agents poll a specific work queue for new work."},null,8,["command"])]),well:o(()=>[e(u)?(s(),t(e(p),{key:0,"work-queue":e(u)},null,8,["work-queue"])):a("",!0)]),default:o(()=>[c(q,{tabs:e(m)},{details:o(()=>[e(u)?(s(),t(e(p),{key:0,"work-queue":e(u)},null,8,["work-queue"])):a("",!0)]),"upcoming-runs":o(()=>[e(u)?(s(),t(e(x),{key:0,"work-queue":e(u)},null,8,["work-queue"])):a("",!0)]),_:1},8,["tabs"])]),_:1})]),_:1})}}});export{S as default};
