import{d as _,u as p,c as a,a as c,w as l,a3 as d,r as k,o as t,j as f,e,a4 as i,h as b,a5 as h,a6 as y,F as D,g as x}from"./index.4d314fbd.js";const w=_({__name:"Blocks",setup(B){const o=p(d.getBlockDocuments),n=a(()=>{var s;return(s=o.response)!=null?s:[]}),u=a(()=>o.executed&&n.value.length==0),r=a(()=>o.executed);return(s,C)=>{const m=k("p-layout-default");return t(),c(m,{class:"blocks"},{header:l(()=>[f(e(i))]),default:l(()=>[e(r)?(t(),b(D,{key:0},[e(u)?(t(),c(e(h),{key:0})):(t(),c(e(y),{key:1,"block-documents":e(n),onDelete:e(o).refresh},null,8,["block-documents","onDelete"]))],64)):x("",!0)]),_:1})}}});export{w as default};
