import{d as m,u as d,c as s,a as i,w as r,ao as _,r as k,o as a,j as y,e,ap as x,h as C,aq as $,ar as g,F as v,g as N}from"./index.4d314fbd.js";const b=m({__name:"Notifications",setup(h){const t=d(_.getNotifications),c=s(()=>{var n;return(n=t.response)!=null?n:[]}),f=s(()=>t.executed&&c.value.length===0),u=s(()=>t.executed);return(n,o)=>{const p=k("p-layout-default");return a(),i(p,{class:"notifications"},{header:r(()=>[y(e(x))]),default:r(()=>[e(u)?(a(),C(v,{key:0},[e(f)?(a(),i(e($),{key:0})):(a(),i(e(g),{key:1,notifications:e(c),onDelete:o[0]||(o[0]=l=>e(t).refresh()),onUpdate:o[1]||(o[1]=l=>e(t).refresh())},null,8,["notifications"]))],64)):N("",!0)]),_:1})}}});export{b as default};
