import{d as _,u as i,c as a,a as n,w as r,C as d,r as w,o as t,j as k,e,D as y,h as x,E as C,G as g,F as h,g as v}from"./index.4d314fbd.js";const B=_({__name:"Flows",setup(F){const u={interval:3e4},s=i(d.getFlows,[{}],u),l=a(()=>{var o;return(o=s.response)!=null?o:[]}),p=a(()=>s.executed&&l.value.length===0),f=a(()=>s.executed);return(o,c)=>{const m=w("p-layout-default");return t(),n(m,{class:"flows"},{header:r(()=>[k(e(y))]),default:r(()=>[e(f)?(t(),x(h,{key:0},[e(p)?(t(),n(e(C),{key:0})):(t(),n(e(g),{key:1,flows:e(l),onDelete:c[0]||(c[0]=$=>e(s).refresh())},null,8,["flows"]))],64)):v("",!0)]),_:1})}}});export{B as default};
