import{d as x,u as C,H as R,c as d,a as n,w as o,b as N,r as s,o as a,e,N as O,g as u,O as y,j as l,z as S,y as $,q as g}from"./index.4d314fbd.js";import{u as B}from"./useRouteParam.0009ec42.js";const j=x({__name:"Deployment",setup(P){const i=B("id"),v=N(),f={interval:3e5},m=C(R.getDeployment,[i.value],f),t=d(()=>m.response),b=d(()=>{var p,r;return(r=(p=t.value)==null?void 0:p.schedule)!=null?r:""});function w(){v.push(g.deployments())}return(p,r)=>{const c=s("p-key-value"),h=s("p-content"),k=s("p-tabs"),D=s("p-layout-well");return a(),n(D,{class:"deployment"},{header:o(()=>[e(t)?(a(),n(e(O),{key:0,deployment:e(t),onUpdate:e(m).refresh,onDelete:w},null,8,["deployment","onUpdate"])):u("",!0)]),well:o(()=>[e(t)?(a(),n(e(y),{key:0,deployment:e(t),alternate:""},null,8,["deployment"])):u("",!0)]),default:o(()=>[e(t)?(a(),n(k,{key:0,tabs:["Overview","Parameters"]},{overview:o(()=>[l(h,{secondary:""},{default:o(()=>{var _;return[l(c,{label:"Schedule",value:e(b)},null,8,["value"]),l(c,{label:"Location",value:e(t).flowData.blob},null,8,["value"]),l(c,{label:"Flow Runner",value:(_=e(t).flowRunner)==null?void 0:_.type},null,8,["value"]),e(S).xl?u("",!0):(a(),n(e(y),{key:0,deployment:e(t)},null,8,["deployment"]))]}),_:1})]),parameters:o(()=>[l(e($),{parameters:e(t).parameters},null,8,["parameters"])]),_:1})):u("",!0)]),_:1})}}});export{j as default};
