import{d as r,u as m,c as k,e as o,a as p,w as t,g as _,b as d,a3 as b,r as i,o as f,j as c,ak as D,al as B,q as h}from"./index.4d314fbd.js";import{u as w}from"./useRouteParam.0009ec42.js";const v=r({__name:"BlockView",setup(y){const s=d(),a=w("blockDocumentId"),n=m(b.getBlockDocument,[a]),e=k(()=>n.response),u=()=>{s.push(h.blocks())};return(C,x)=>{const l=i("p-layout-default");return o(e)?(f(),p(l,{key:0,class:"block-view"},{header:t(()=>[c(o(D),{"block-document":o(e),onDelete:u},null,8,["block-document"])]),default:t(()=>[c(o(B),{"block-document":o(e)},null,8,["block-document"])]),_:1})):_("",!0)}}});export{v as default};
