import{d as _,u as y,c as a,a as l,w as r,H as f,r as k,o as n,j as v,e,L as x,h as $,M as g,K as h,F as C,g as b}from"./index.4d314fbd.js";const w=_({__name:"Deployments",setup(B){const m={interval:3e4},t=y(f.getDeployments,[{}],m),p=a(()=>{var o;return(o=t.response)!=null?o:[]}),c=a(()=>t.executed&&p.value.length===0),d=a(()=>t.executed);return(o,s)=>{const u=k("p-layout-default");return n(),l(u,{class:"deployments"},{header:r(()=>[v(e(x))]),default:r(()=>[e(d)?(n(),$(C,{key:0},[e(c)?(n(),l(e(g),{key:0})):(n(),l(e(h),{key:1,deployments:e(p),onUpdate:s[0]||(s[0]=i=>e(t).refresh()),onDelete:s[1]||(s[1]=i=>e(t).refresh())},null,8,["deployments"]))],64)):b("",!0)]),_:1})}}});export{w as default};
